//
//	attitudelocator.cc: test the AttitudeLoator class
//
//	DESCRIPTION:
//		test the AttitudeLocator class with the AHF created
//		by ../odf/ahfgen. In this file the Ra/Dec/Pos changes
//		over 235 min linearly from (30./-20/-99) -> (180/0/0)
//		The latter attitude corresponds to a ONE matrix.
//

#include "OalServer.h"

#include <iostream>
#include <cassert>
#include <cmath>
#include "errstr.h"
#include "STime.h"
#include "Celestial.h"

double t(const STime &tp)
{
	return tp - Xmm::missionReferenceTime;
}

const double period = 235. * 60.;
double f(double min, double max, double t)
{
    return DEG_TO_RAD* (min + (max-min)/period * t);
}


int main()
{
	const double ra0 = 30., ra1 = 180.,
				 dec0 = -20., dec1 = 0.,
				 pos0 = -99., pos1 = 0.;
	STime startAttData("1999-08-28T08:05:00");

	const Odf *odf = oalServer.newOdf("../odf");
	errstr<<*odf<<Sas::message(Msg::App, Msg::Verbose);

	STime startExp(odf->exposure(Xmm::EMOS1, Xmm::IMAGING, 1)->
					actualStartTime()),
		  endExp(odf->exposure(Xmm::EMOS1, Xmm::IMAGING, 1)->
					actualEndTime());

	double diff = .5 *(endExp - startExp);
	SpacecraftAttitude att;
	cout<<"Exposure time frame: "<<startExp<<" - "<<endExp<<endl;
	const AttitudeLocator *attloc = oalServer.attitudeLocator();

	//
	//	check attitude at beginning of exposure
	//
	double tt;
	try {
		att = attloc->attitude(t(startExp));
		cout.precision(8);
			cout<<"Attitude at start of exposure at: "<<startExp<<" ("
				<<t(startExp)<<"):\n"<<att<<endl;
		cout<<"["<<att.ra()<<"/"<<att.dec()<<"/"<<att.apos()<<"]\n";
		tt = startExp - startAttData - 7.;
			//	exposure starts 7 seconds after data

		assert(fabs(att.ra()/f(ra0, ra1, tt)-1.) < 1.e-5);
		assert(fabs(att.dec()/f(dec0, dec1, tt)-1.) < 1.e-5);
		assert(fabs(att.apos()/f(pos0, pos1, tt)-1.) < 1.e-5);

		assert(fabs(att.ra()/f(ra0, ra1, tt+7)-1.) < 1.e-2);
		assert(fabs(att.dec()/f(dec0, dec1, tt+7)-1.) < 1.e-2);
		assert(fabs(att.apos()/f(pos0, pos1, tt+7)-1.) < 1.e-2);
	} catch(...) {
		cerr<<"No position data available for t="<<startExp<<endl;
		return 1;
	}


	//
	//	check attitude in middle of exposure
	//
	STime midExp(startExp + diff);
	try {
		att = attloc->attitude(t(midExp));
	} catch(...) {
		cout<<"No data\n"; return 1;
	}
	cout<<"Attitude in middle of exposure at: "<<midExp<<" ("
		<<t(midExp)<<"):\n"<<att<<endl;
	cout<<"["<<att.ra()<<"/"<<att.dec()<<"/"<<att.apos()<<"]\n";
	tt = startExp - startAttData - 7. + diff;
		//	exposure starts 7 seconds after data

	assert(fabs(att.ra()/f(ra0, ra1, tt)-1.) < 1.e-5);
	assert(fabs(att.dec()/f(dec0, dec1, tt)-1.) < 1.e-5);
	assert(fabs(att.apos()/f(pos0, pos1, tt)-1.) < 1.e-5);
	

	//
	//	check attitude at end of exposure
	//
	try {
		att = attloc->attitude(t(endExp));
	} catch(...) {
		cerr<<"No data\n"; return 1;
	}
	AttitudeMatrix attmat(att.toAttitudeMatrix());
	cout<<"Attitude at end of exposure at: "<<endExp<<" ("
		<<t(endExp)<<") :\n"<<att<<endl;
	cout<<"["<<att.ra()<<"/"<<att.dec()<<"/"<<att.apos()<<"]\n";
	cout<<"The corresponding attitude matrix is:\n"<<attmat<<endl;

	tt = endExp - startAttData - 7.;
		//	exposure starts 7 seconds after data

	assert(fabs(att.ra()/f(ra0, ra1, tt)-1.) < 1.e-5);
	assert(fabs(att.dec()) < 1.e-10);	// Decl and position angle are 0 at
	assert(fabs(att.apos()) < 1.e-10); // the end

	//
	//	check for unity matrix: all Euler angles must be 0
	//
	EulerAngles eul = attmat.toEulerAngles();	
	assert(fabs(eul.phi() * RAD_TO_DEG) < 1.e-10);
	assert(fabs(eul.theta() * RAD_TO_DEG) < 1.e-10);
	assert(fabs(eul.psi() * RAD_TO_DEG) < 1.e-10);

	//
	//	check attitude in a gap - this should fail with an exception
	//
	STime mid2Exp(midExp - 2. * 60);
	cout<<"Trying to retrieve attitude in a gap at: "<<mid2Exp<<" ... ";
	try {
		att = attloc->attitude(t(mid2Exp));
		cout<<"This should fail - but it did not\n"; return 1;
	} catch(...) {
		cout<<"ok, failed\n";
	}

	cout<<"\nAll tests successfully passed.\n";
	return 0;
}
