//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeAxis.cc	--- implementation TimeAxis classes
//
//	DESCRIPTION:
//		Implementation of TimeAxis classes
//
//	CHANGE HISTORY:
//		UL, 1998-08-17: 1.0.0: - first version
//		UL, 1999-03-19: 1.1.0: - several bug fixes causing boundary cases
//								 to be treated incorrectly
//		UL, 1999-08-05: 1.2.0: - more explicit error message in TimeAxisIndex
//								 constructor
//		UL, 2001-03-28: 1.3.0: - bug fix: TimeAxisIndex::operator++() did
//								 not work correctly when duration of
//								 axis bin was 0
//		UL, 2001-04-04: 1.4.0: - fixed another problem caused by consecutive
//								 bins of duration 0s
//
//	AUTHOR:
//		U. Lammers, Mon Aug 17 13:58:48 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files (ODF-ICD), K. Galloway
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/oal/src/TimeAxis.cc,v 1.11 2003/06/16 08:41:35 gvacanti Exp $
//

#include <algorithm>
#include <cmath>

#include "TimeAxis.h"
#include "errstr.h"
#include "common.h"

//
//	---( TimeAxis )--------------------------------------------------------
//
ostream &operator<<(ostream &os, const TimeAxis &tax)
{
	os<<"TimeAxis(size(time)="<<tax._time.size()<<", granularity="
	  <<tax._granularity<<")";

	return os;
}


//
//	---( TimeAxisIndex )---------------------------------------------------
//
TimeAxisIndex::TimeAxisIndex(const TimeAxis *tAxis, double timeVal)
		: _majorInd(0), _minorInd(0), _valid(false),
		  _whatsWrong(NOTHING), _tAxis(tAxis)
{
	unsigned nMax = tAxis->_time.size(),
			 n = 0;

	if (!nMax)
		errstr<<"nMax == 0: This should not happen"
			  <<Sas::error("ThisShouldNotHappen");

	//
	//	find first point which is larger than <timeVal>
	//
	while (n < nMax && tAxis->_time[n] <= timeVal)
			++n;
	if (n > 0) {
		_valid = true;
		_whatsWrong = NOTHING;
		_majorInd = n-1;
		_minorInd = tAxis->_time[n-1] + tAxis->_duration[n-1] > timeVal
					 ? INT((timeVal - tAxis->_time[n-1])/tAxis->_granularity)
					 : max(INT(tAxis->_duration[n-1]/tAxis->_granularity +.5)
						   - 1, 0);
	} else
		_whatsWrong = MINREACHED;
}


TimeAxisIndex &TimeAxisIndex::operator++()
{
	if (_valid) {
		int maxInd = max(INT(_tAxis->_duration[_majorInd]/
							 _tAxis->_granularity + .5) - 1, 0);
		if (_minorInd < maxInd)
			++_minorInd;
		else
			if (_majorInd < INT(_tAxis->_time.size())-1) {
				++_majorInd;
				_minorInd = 0;
			} else {
				_valid = false;
				_whatsWrong = MAXREACHED;
			}
	} else
		if (_whatsWrong == MINREACHED) {
			_majorInd = _minorInd = 0;
			_valid = true;
			_whatsWrong = NOTHING;
		}

	return *this;
} // TimeAxisIndex::operator++


TimeAxisIndex &TimeAxisIndex::operator--()
{
	if (_valid)
		if (_minorInd > 0)
			--_minorInd;
		else
			if (_majorInd > 0) {
				--_majorInd;
				_minorInd = max(INT(_tAxis->_duration[_majorInd]/
										_tAxis->_granularity + .5) - 1, 0);
			} else {
				_valid = false;
				_whatsWrong = MINREACHED;
			}
	else
		if (_whatsWrong == MAXREACHED) {
			_majorInd = _tAxis->_time.size()-1;
			_minorInd = max(INT(_tAxis->_duration[_majorInd]/
										_tAxis->_granularity + .5) - 1, 0);
			_valid = true;
			_whatsWrong = NOTHING;
		}

	return *this;
} // TimeAxisIndex::operator--


double TimeAxisIndex::distance(double timeVal) const
{
	if (_valid) {
		return _tAxis->_time[_majorInd] +
					_tAxis->_granularity * _minorInd - timeVal;
	} else if (_whatsWrong == MINREACHED)
		return -HUGE_VAL;
	else if (_whatsWrong == MAXREACHED)
		return HUGE_VAL;
	else
		errstr<<"This cannot happen"<<Sas::fatal("TimeAxisIndex::distance");

	return 0.;
} // TimeAxisIndex::distance()


ostream &operator<<(ostream &os, const TimeAxisIndex &tax)
{
	os<<"TimeAxisIndex(majorInd="<<tax.majorInd()<<", minorInd="
	  <<tax.minorInd()<<", valid="<<tax.valid()<<", whatsWrong="
	  <<UNT(tax.whatsWrong())<<")";

	return os;
}


